/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include <sys/types.h>
#include <regex.h>
#include <ctype.h>
#include <mailutils/mailutils.h>
#include <libmilter/mfapi.h>

#ifndef MAXPACKET
# define MAXPACKET 8192	/* max packet size used internally by BIND */
#endif 

typedef enum {
	getmx_success,
	getmx_not_found,
	getmx_failure,
	getmx_temp_failure,
} getmx_status;

#define getmx_resolved(c) ((c) == getmx_success || (c) == getmx_not_found)

#define MAXMXCOUNT 32
typedef char *mxbuf_t[MAXMXCOUNT];

#define DEFAULT_PIDFILE MAILFROMSTATEDIR "/mailfromd.pid"
#define DEFAULT_DATABASE MAILFROMSTATEDIR "/mailfromd"
#define DEFAULT_CONFIG_FILE SYSCONFDIR "/mailfromd.rc"

#define __DBG(lev) if (debug_level >= lev)

#define debug(lev,text) \
 __DBG(lev) debug_log("%s:%lu:%s: %s", __FILE__, __LINE__, __FUNCTION__, text)
#define debug1(lev,fmt,v) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, __FILE__, __LINE__, __FUNCTION__, v)
#define debug2(lev,fmt,a,b) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, \
   __FILE__, __LINE__, __FUNCTION__, a, b)
#define debug3(lev,fmt,a,b,c) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, \
   __FILE__, __LINE__, __FUNCTION__, a, b, c)
#define debug4(lev,fmt,a,b,c,d) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, \
   __FILE__, __LINE__, __FUNCTION__, a, b, c, d)
#define debug5(lev,fmt,a,b,c,d,e) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, \
   __FILE__, __LINE__, __FUNCTION__, a, b, c, d, e)
#define debug6(lev,fmt,a,b,c,d,e,f) \
 __DBG(lev) debug_log("%s:%lu:%s: " fmt, \
   __FILE__, __LINE__, __FUNCTION__, a, b, c, d, e, f)

extern int debug_level;
extern time_t expire_interval;
extern char *smtp_domain;
extern char *postmaster_email;


/* Configuration file support */

typedef struct node NODE;

struct locus {
	char *file;
	size_t line;
};

struct if_node {
	NODE *cond;
	NODE *if_true;
	NODE *if_false;
};

struct poll_action {
	struct poll_action *next;
	struct locus locus;
	getmx_status status;
	NODE *node;
};

struct poll_node {
	NODE *email;
	NODE *client_addr;
	NODE *ehlo;
	NODE *mailfrom;
	struct poll_action *actions;
};

enum bin_opcode {
	bin_and,
	bin_or,
	bin_eq,
	bin_ne,
	bin_regex,
	bin_match,
	bin_fnmatch
};

struct bin_node {
	enum bin_opcode opcode;
	NODE *arg[2];
};

struct regex_node { /* A special form of bin_node */
	enum bin_opcode opcode; /* always == bin_regex */
	NODE *arg;
	regex_t regex;
};

enum un_opcode {
	unary_not
};

struct un_node {
	enum un_opcode opcode;
	NODE *arg;
};

struct return_node {
	sfsistat stat;
	char *code;
	char *xcode;
	char *message;
};

enum header_opcode {
	header_add,
	header_replace,
	header_delete
};

struct header_node {
	enum header_opcode opcode;
	char *name;
	char *value;
};

enum node_type {
	node_type_string,
	node_type_symbol,
	node_type_if,
	node_type_poll,
	node_type_bin,
	node_type_un,
	node_type_return,
	node_type_header,
	node_type_hostname,
	node_type_relayed_domain
};

struct node {
	NODE *next;
	enum node_type type;
	struct locus locus;
	union {
		char *string;
		struct if_node cond;
		struct poll_node poll;
		struct bin_node bin;
		struct un_node un;
		struct return_node ret;
		struct header_node hdr;
		struct regex_node re;
		NODE *node;
	} v;
};

void debug_log(char *fmt, ...);
void parse_error(char *fmt, ...);
int yyparse();
int source(char *, int);
int parse_config(char *name, int ydebug, int ldebug);
const struct locus *get_locus(void);
void trace(const char *fmt, ...);

const char *header_command_str(enum header_opcode opcode);
const char *sfsistat_str(sfsistat stat);
int priv_store_header_command(SMFICTX *ctx, struct header_node *node);

getmx_status getmx(char *ipstr, mxbuf_t mxbuf);

void dict_init(list_t *dict);
char *dict_install(list_t dict, char *name, char *value);
void dict_destroy(list_t *dict);
char *dict_getsym(void *data, char *str);
int set_option(char *name, char *value, int override);
sfsistat run_program(SMFICTX *ctx);
sfsistat test_program(list_t dict);
void print_config(void);


/* DBM support */
#include "mu_dbm.h"

#ifdef USE_DBM
# define cache_get _mailfrom_cache_get
# define cache_insert _mailfrom_cache_insert
# define cache_get2 _mailfrom_cache_get2
# define cache_insert2 _mailfrom_cache_insert2
# define cache_delete _mailfrom_cache_delete
# define cache_list_item _mailfrom_cache_list_item
# define cache_list_db _mailfrom_cache_list_db
# define cache_expire_db _mailfrom_cache_expire_db
getmx_status cache_get(char *email);
void cache_insert(char *email, getmx_status rc);
void cache_delete(char *email);
getmx_status cache_get2(char *email, char *client_addr);
void cache_insert2(char *email, char *client_addr, getmx_status rc);
void cache_list_item(char *email);
void cache_list_db(void);
void cache_expire_db(void);
#else
# define cache_get(email) getmx_failure
# define cache_insert(email, rc)
# define cache_get2(email, client_addr) getmx_failure
# define cache_insert2(email, client_addr, rc)
# define cache_delete(email)
# define cache_list_item(email)
# define cache_list_db()
# define cache_expire_db()
#endif


